#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

glance_api_group = cfg.OptGroup(name="glance_api",
                                title="Glance API Service Options")

GlanceAPIGroup = [cfg.BoolOpt("system_scope_supported",
                              default=False,
                              help="Whether or not system scope "
                                   "is supported")]
